from typing_extensions import Final  # noqa: F401

MAILBOX_DOMAIN = 'lokole.ca'  # type: Final
RECEIVED_FOLDER = 'received'  # type: Final
SENT_FOLDER = 'sent'  # type: Final
FUTURE_TIMESTAMP = 2100000000  # type: Final
